#!/usr/bin/env/python
"""
Usage:
    chem_tensorflow_dense.py [options]

Options:
    -h --help                Show this screen.
    --config-file FILE       Hyperparameter configuration file path (in JSON format)
    --config CONFIG          Hyperparameter configuration dictionary (in JSON format)
    --log_dir NAME           log dir name
    --data_dir NAME          data dir name
    --restore FILE           File to restore weights from.
    --freeze-graph-model     Freeze weights of graph model components.
    --evaluate               example evaluation mode using a restored model
"""

from typing import Sequence, Any
from collections import defaultdict
import numpy as np
import tensorflow as tf
import sys, traceback
import pdb
import json

from utils.utils import glorot_init



'''
Comments provide the expected tensor shapes where helpful.

Key to symbols in comments:
---------------------------
[...]:  a tensor
; ; :   a list
b:      batch size
e:      number of edge types (4)
v:      number of vertices per graph in this batch
h:      GNN hidden size
'''

class DenseGGNNModel():
    def __init__(self, opt):
        self.node_dim = opt.node_dim
        self.hidden_layer_size = opt.hidden_layer_size
        self.num_hidden_layer = opt.num_hidden_layer
        self.aggregation_type = opt.aggregation
        self.distributed_function = opt.distributed_function
        self.num_labels = opt.n_classes
        self.num_edge_types = opt.n_edge_types
        self.num_timesteps= opt.n_steps
        self.placeholders = {}
        self.weights = {}

        self.prepare_specific_graph_model()
        self.nodes_representation = self.compute_nodes_representation()
        # self.graph_representation = self.pooling_layer(self.nodes_representation)

        graph_representation, attention_scores = self.aggregation_layer(self.nodes_representation, self.aggregation_type, self.distributed_function)

        self.graph_representation = graph_representation
        self.attention_scores = attention_scores
        
        if self.num_hidden_layer == 1:
            self.logits = self.hidden_layer(self.graph_representation, self.node_dim, self.num_labels)
        else:
            self.logits = self.hidden_layer(self.graph_representation, self.node_dim, self.hidden_layer_size)
            for i in range(1, self.num_hidden_layer):
                self.logits = self.hidden_layer(self.logits, self.hidden_layer_size, self.num_labels)

        self.loss = self.loss_layer(self.logits)
        self.softmax_values = self.softmax(self.logits)


    def prepare_specific_graph_model(self) -> None:
        node_dim = self.node_dim

        # initializer = tf.contrib.layers.xavier_initializer()
        # inputs
        # self.placeholders['graph_state_keep_prob'] = tf.placeholder(tf.float32, None, name='graph_state_keep_prob')
        # self.placeholders['edge_weight_dropout_keep_prob'] = tf.placeholder(tf.float32, None, name='edge_weight_dropout_keep_prob')
        self.placeholders['initial_node_representation'] = tf.placeholder(tf.float32, [None, None, self.node_dim], name='node_features')
        self.placeholders['num_vertices'] = tf.placeholder(tf.int32, (),  name='num_vertices')
        self.placeholders['labels'] = tf.placeholder(tf.int32, (None, self.num_labels))

        self.placeholders['adjacency_matrix'] = tf.placeholder(tf.float32,[None, self.num_edge_types, None, None], name='adjacency_matrix')     # [b, e, v, v]
        self.__adjacency_matrix = tf.transpose(self.placeholders['adjacency_matrix'], [1, 0, 2, 3])                    # [e, b, v, v]

        # weights
        self.weights['edge_weights'] = tf.Variable(glorot_init([self.num_edge_types, node_dim, node_dim]),name='edge_weights')
        self.weights['edge_biases'] = tf.Variable(glorot_init([self.num_edge_types, 1, node_dim]).astype(np.float32),name='edge_biases')
        
        self.xavier_initializer = tf.contrib.layers.xavier_initializer()
        # self.weights["hidden_layer_weights"] = tf.Variable(xavier_initializer([self.node_dim, self.num_labels]), name='hidden_layer_weights')
        # self.weights["hidden_layer_biases"] = tf.Variable(xavier_initializer([self.num_labels,]), name='hidden_layer_biases')

        self.weights['attention_weights'] = tf.Variable(glorot_init([self.node_dim,1]).astype(np.float32),name='attention_weights')
        

        with tf.variable_scope("gru_scope"):
            cell = tf.contrib.rnn.GRUCell(node_dim)
            # cell = tf.nn.rnn_cell.DropoutWrapper(cell, state_keep_prob=self.placeholders['graph_state_keep_prob'])
            self.weights['node_gru'] = cell

    def compute_nodes_representation(self):
        node_dim = self.node_dim
        v = self.placeholders['num_vertices']
        h = self.placeholders['initial_node_representation']                                                # [b, v, h]
        h = tf.reshape(h, [-1, node_dim])

        with tf.variable_scope("gru_scope") as scope:
            for i in range(self.num_timesteps):
                if i > 0:
                    tf.get_variable_scope().reuse_variables()
                for edge_type in range(self.num_edge_types):
                    # print("edge type : " + str(edge_type))
                    m = tf.matmul(h, self.weights['edge_weights'][edge_type])                               # [b*v, h]

                    m = tf.reshape(m, [-1, v, node_dim])                                                       # [b, v, h]
                    m += self.weights['edge_biases'][edge_type]                                             # [b, v, h]

                    if edge_type == 0:
                        acts = tf.matmul(self.__adjacency_matrix[edge_type], m)
                    else:
                        acts += tf.matmul(self.__adjacency_matrix[edge_type], m)
                acts = tf.reshape(acts, [-1, node_dim])                                                        # [b*v, h]

                h = self.weights['node_gru'](acts, h)[1]                                                    # [b*v, h]
            last_h = tf.reshape(h, [-1, v, node_dim])
        return last_h

    def pooling_layer(self, nodes_representation):
        """Creates a max dynamic pooling layer from the nodes."""
        with tf.name_scope("pooling"):
            pooled = tf.reduce_max(nodes_representation, axis=1)
            return pooled

    def hidden_layer(self, pooled, input_size, output_size):
        """Create a hidden feedforward layer."""
        # self.weights["hidden_layer_weights"] = tf.Variable(xavier_initializer([self.node_dim, self.num_labels]), name='hidden_layer_weights')
        # self.weights["hidden_layer_biases"] = tf.Variable(xavier_initializer([self.num_labels,]), name='hidden_layer_biases')

        with tf.name_scope("hidden"):
            weights = tf.Variable(self.xavier_initializer([input_size, output_size]))
            biases = tf.Variable(self.xavier_initializer([output_size,]))
            return tf.nn.leaky_relu(tf.matmul(pooled, weights) + biases)

    def loss_layer(self, logits_node):
        """Create a loss layer for training."""
        labels = self.placeholders['labels']

        with tf.name_scope('loss_layer'):
            cross_entropy = tf.nn.softmax_cross_entropy_with_logits(
                labels=labels, logits=logits_node, name='cross_entropy'
            )

            loss = tf.reduce_mean(cross_entropy, name='cross_entropy_mean')
            return loss

    def aggregation_layer(self, nodes_representation, aggregation_type, distributed_function):
        # nodes_representation is (batch_size, max_graph_size, self.node_dim)
        w_attention = self.weights['attention_weights']
        with tf.name_scope("global_attention"):
            batch_size = tf.shape(nodes_representation)[0]
            max_tree_size = tf.shape(nodes_representation)[1]

            # (batch_size * max_graph_size, self.node_dim)
            flat_nodes_representation = tf.reshape(nodes_representation, [-1, self.node_dim])
            aggregated_vector = tf.matmul(flat_nodes_representation, w_attention)

            attention_score = tf.reshape(aggregated_vector, [-1, max_tree_size, 1])

            """A note here: softmax will distributed the weights to all of the nodes (sum of node weghts = 1),
            an interesting finding is that for some nodes, the attention score will be very very small, i.e e-12, 
            thus making parts of aggregated vector becomes near zero and affect on the learning (very slow nodes_representationergence
            - Better to use sigmoid"""

            if distributed_function == 0:
                attention_weights = tf.nn.softmax(attention_score, dim=1)
            if distributed_function == 1:
                attention_weights = tf.nn.sigmoid(attention_score)

            # TODO: reduce_max vs reduce_sum vs reduce_mean
            if aggregation_type == 1:
                print("Using tf.reduce_sum...........")
                weighted_average_nodes = tf.reduce_sum(tf.multiply(nodes_representation, attention_weights), axis=1)
            if aggregation_type == 2:
                print("Using tf.reduce_max...........")
                weighted_average_nodes = tf.reduce_max(tf.multiply(nodes_representation, attention_weights), axis=1)
            if aggregation_type == 3:
                print("Using tf.reduce_mean...........")
                weighted_average_nodes = tf.reduce_mean(tf.multiply(nodes_representation, attention_weights), axis=1)

            return weighted_average_nodes, attention_weights


    def softmax(self, logits_node):
        """Apply softmax to the output layer."""
        with tf.name_scope('output'):
            return tf.nn.softmax(logits_node)



#!/usr/bin/env/python

import numpy as np
import tensorflow as tf
import queue
import threading
import numpy as np
from sklearn import preprocessing
import os, itertools

SMALL_NUMBER = 1e-7


def glorot_init(shape):
    initialization_range = np.sqrt(6.0 / (shape[-2] + shape[-1]))
    return np.random.uniform(low=-initialization_range, high=initialization_range, size=shape).astype(np.float32)

def scale_attention_score(attention_array, top_node):
    attention_array = np.array(attention_array)
  
    top_n_idx = np.argsort(attention_array)[-top_node:][::-1]
    top_n_values = [attention_array[i] for i in top_n_idx]

    max_point = max(top_n_values)
    min_point = min(top_n_values)

    norm_data = []
    for i, point in enumerate(attention_array):
        if i in top_n_idx:
            point_norm = float((point - min_point)/(max_point - min_point))

            # Means to scale from 0 to 1
            point_scaled = point_norm * (1.0 - 0.0) + 0.0
            norm_data.append(point_scaled)
        else:
            norm_data.append(0.0)

    return norm_data


def scale_attention_score_by_group(attention_array):

    attention_array = np.array(attention_array)
    clusters = [list(g) for _, g in itertools.groupby(attention_array, lambda x: x)]
    average = float(1/len(clusters))

    scaled_attention_array = []
    for score in attention_array:
        for i, cluster in enumerate(clusters):
            if score in cluster:
                temp_score = average * (len(clusters) - i)
                scaled_attention_array.append(round(temp_score,5))

    return scaled_attention_array



class ThreadedIterator:
    """An iterator object that computes its elements in a parallel thread to be ready to be consumed.
    The iterator should *not* return None"""

    def __init__(self, original_iterator, max_queue_size: int=2):
        self.__queue = queue.Queue(maxsize=max_queue_size)
        self.__thread = threading.Thread(target=lambda: self.worker(original_iterator))
        self.__thread.start()

    def worker(self, original_iterator):
        for element in original_iterator:
            assert element is not None, 'By convention, iterator elements much not be None'
            self.__queue.put(element, block=True)
        self.__queue.put(None, block=True)

    def __iter__(self):
        next_element = self.__queue.get(block=True)
        while next_element is not None:
            yield next_element
            next_element = self.__queue.get(block=True)
        self.__thread.join()


class MLP(object):
    def __init__(self, in_size, out_size, hid_sizes, dropout_keep_prob):
        self.in_size = in_size
        self.out_size = out_size
        self.hid_sizes = hid_sizes
        self.dropout_keep_prob = dropout_keep_prob
        self.params = self.make_network_params()

    def make_network_params(self):
        dims = [self.in_size] + self.hid_sizes + [self.out_size]
        weight_sizes = list(zip(dims[:-1], dims[1:]))
        weights = [tf.Variable(self.init_weights(s), name='MLP_W_layer%i' % i)
                   for (i, s) in enumerate(weight_sizes)]
        biases = [tf.Variable(np.zeros(s[-1]).astype(np.float32), name='MLP_b_layer%i' % i)
                  for (i, s) in enumerate(weight_sizes)]

        network_params = {
            "weights": weights,
            "biases": biases,
        }

        return network_params

    def init_weights(self, shape):
        return np.sqrt(6.0 / (shape[-2] + shape[-1])) * (2 * np.random.rand(*shape).astype(np.float32) - 1)

    def __call__(self, inputs):
        acts = inputs
        for W, b in zip(self.params["weights"], self.params["biases"]):
            hid = tf.matmul(acts, tf.nn.dropout(W, self.dropout_keep_prob)) + b
            acts = tf.nn.relu(hid)
        last_hidden = hid
        return last_hidden


def calculate_performance(y_actual, y_hat):
    TP = 0
    FP = 0
    TN = 0
    FN = 0

    for i in range(len(y_hat)): 
        if y_hat[i] == y_actual[i] ==1:
            TP += 1
        if y_hat[i] == 1 and y_actual[i] == 0:
            FP += 1
        if y_hat[i] == y_actual[i] == 0:
            TN += 1
        if y_hat[i] == 0 and y_actual[i] == 1:
            FN += 1

    return (TP, FP, TN, FN)



import numpy as np
import os
from os import listdir
from os.path import isfile, join
import collections
import re
from tqdm import trange
from tqdm import *
import random
#import pickle
import pyarrow
from collections import defaultdict

def load_graphs_from_file(file_name):
    data_list = []
    edge_list = []
    target_list = []
    
    with open(file_name,'r') as f:
        for line in f:
            if len(line.strip()) == 0:
                data_list.append([edge_list,target_list])
                edge_list = []
                target_list = []
            else:
                digits = []
                line_tokens = line.split(" ")

                if line_tokens[0] == "?":
                    for i in range(1, len(line_tokens)):
                        
                        digits.append(int(line_tokens[i]))
                    target_list.append(digits)
                else:
                    for i in range(len(line_tokens)):
                        digits.append(int(line_tokens[i]))
                    edge_list.append(digits)
    # print(data_list)
    return data_list


def lookup_vector(node_type, embeddings):
    # look_up_vector = embed_loopkup[node_type]

    nodes.append(embeddings[int(n)])


def load_program_graphs_from_directory(directory, is_training=True, is_testing=True, n_classes=3):

    node_id_data_list = []
    node_type_data_list = []
    if is_training == True:
        dir_path =  os.path.join(directory,"train")
    if is_testing == True:
        dir_path =  os.path.join(directory,"test")
    filenames = []
    for f in listdir(dir_path):
      if isfile(join(dir_path, f)):
         filenames.append(f)
    int_filenames = [int(re.search('_(.*).txt', x).group(1)) for x in filenames]
    ordered_filenames = sorted(int_filenames)
    lookup = {}
    for i in range(1, 1+len(ordered_filenames)):
        if is_training:
           lookup[i] = join(dir_path, "train_%s.txt" % str(ordered_filenames[i-1]))
        if is_testing:
           lookup[i] = join(dir_path, "test_%s.txt" % str(ordered_filenames[i-1]))
    for i in trange(1, 1+n_classes):
        path = lookup[i]
        # print(path)
        label = i
        node_id_data_list_class_i = []
        node_type_data_list_class_i = []
        node_id_edge_list_class_i = []
        node_type_edge_list_class_i = []
        target_list_class_i = []

        # print("--------------------------")
        with open(path,'r') as f:
            for line in f: 
                if len(line.strip()) == 0:
                    # print(edge_list_class_i)
                    node_id_data_list_class_i.append([node_id_edge_list_class_i,target_list_class_i])
                    node_type_data_list_class_i.append([node_type_edge_list_class_i,target_list_class_i])
                    node_id_edge_list_class_i = []
                    node_type_edge_list_class_i = []
                    target_list_class_i = []
                else:
                    node_id_digits = []
                    node_type_digits = []
                    line_tokens = line.split(" ")
                    
                    if line_tokens[0] == "?":
                        # print(label)
                        target_list_class_i.append([label])
                    else:

                        for j in range(len(line_tokens)):
                            if "," in line_tokens[j]:
                                splits = line_tokens[j].split(",")
                                node_id = splits[0]
                                node_type = splits[1]
                                node_id_digits.append(int(node_id))
                                node_type_digits.append(int(node_type))
                            else:
                                node_id_digits.append(int(line_tokens[j]))
                                node_type_digits.append(int(line_tokens[j]))

                        node_id_edge_list_class_i.append(node_id_digits)
                        node_type_edge_list_class_i.append(node_type_digits)

        # if data_percentage < 1.0:
        #     print("Cutting down " + str(data_percentage) + " of all data......")
        #     slicing = int(len(node_id_data_list_class_i)*data_percentage)
        #     print("Remaining data : " + str(slicing) + "......")
        #     node_id_data_list_class_i = node_id_data_list_class_i[:slicing]


        node_id_data_list.extend(node_id_data_list_class_i)
        node_type_data_list.extend(node_type_data_list_class_i)

    return node_id_data_list, node_type_data_list

def load_single_program(path):
  
    node_id_data_list_class_i = []
    node_type_data_list_class_i = []
    node_id_edge_list_class_i = []
    node_type_edge_list_class_i = []
    target_list_class_i = []

    # print("--------------------------")
    with open(path,'r') as f:
        for line in f: 
            if len(line.strip()) == 0:
                # print(edge_list_class_i)
                node_id_data_list_class_i.append([node_id_edge_list_class_i,target_list_class_i])
                node_type_data_list_class_i.append([node_type_edge_list_class_i,target_list_class_i])
                node_id_edge_list_class_i = []
                node_type_edge_list_class_i = []
                target_list_class_i = []
            else:
                node_id_digits = []
                node_type_digits = []
                line_tokens = line.split(" ")
                
                if line_tokens[0] == "?":
                    # print(label)
                    splits = line_tokens[2].split("/")
                    # print(splits)
                    try: 
                       label = int(splits[len(splits)-2])
                    except:
                       label = 1

                    target_list_class_i.append([label])
                else:

                    for j in range(len(line_tokens)):
                        if "," in line_tokens[j]:
                            splits = line_tokens[j].split(",")
                            node_id = splits[0]
                            node_type = splits[1]
                            node_id_digits.append(int(node_id))
                            node_type_digits.append(int(node_type))
                        else:
                            node_id_digits.append(int(line_tokens[j]))
                            node_type_digits.append(int(line_tokens[j]))

                    node_id_edge_list_class_i.append(node_id_digits)
                    node_type_edge_list_class_i.append(node_type_digits)
    return node_id_data_list_class_i, node_type_data_list_class_i

def find_max_edge_id(data_list):
    max_edge_id = 0
    for data in data_list:
        edges = data[0]
        for item in edges:
            if item[1] > max_edge_id:
                max_edge_id = item[1]
    return max_edge_id

def find_max_node_id(data_list):
    max_node_id = 0
    for data in data_list:
        edges = data[0]
        for item in edges:

            if item[0] > max_node_id:
                max_node_id = item[0]
            if item[2] > max_node_id:
                max_node_id = item[2]
    return max_node_id

def find_min_node_id(data_list):
    min_node_id = 0
    for data in data_list:
        edges = data[0]
        for item in edges:

            if item[0] < min_node_id:
                min_node_id = item[0]
            if item[2] < min_node_id:
                min_node_id = item[2]
    return min_node_id
    # return 48

def find_num_nodes_of_graph(graph):
    max_node_id = 0
    for item in graph:
        if item[0] > max_node_id:
            max_node_id = item[0]
        if item[2] > max_node_id:
            max_node_id = item[2]
    return max_node_id

def split_set(data_list,num):
    n_examples = len(data_list)
    idx = range(n_examples)
    train = idx[:num]
    val = idx[-num:]
    return np.array(data_list)[train],np.array(data_list)[val]

def split_set_by_percentage(data_list,percentage):
    n_examples = len(data_list)
    train_num = int(n_examples * percentage)

    idx = range(n_examples)
    train = idx[:train_num]
    val = idx[train_num:n_examples]
    return np.array(data_list)[train],np.array(data_list)[val]


def convert_program_data(data_list):
    # n_nodes = find_max_node_id(data_list)
    class_data_list = []
 
    for item in data_list:
        edge_list = item[0]
        target_list = item[1]
        for target in target_list:
            task_type = target[0]
            task_output = target[-1]
            class_data_list.append([edge_list, task_output])
    return class_data_list


def create_adjacency_matrix(edges, n_nodes, n_edge_types):
    a = np.zeros([n_nodes, n_nodes * n_edge_types * 2])
    for edge in edges:

        src_idx = edge[0]
        e_type = edge[1]
        tgt_idx = edge[2]
       
        if tgt_idx < len(a):
           a[tgt_idx-1][(e_type - 1) * n_nodes + src_idx - 1] =  1
        if src_idx <len(a):
           a[src_idx-1][(e_type - 1 + n_edge_types) * n_nodes + tgt_idx - 1] =  1
    return a

def create_embedding_matrix(node_id_edges, node_type_edges, n_nodes, pretrained_embeddings):
    a = np.zeros([n_nodes, 30])
    # print(a.shape)
    for i in range(len(node_id_edges)):
        node_type = node_type_edges[i][0]
        # print(node_type)
        src_idx = node_id_edges[i][0]
        a[src_idx-1] = pretrained_embeddings[int(node_type)]
    return a


def graph_to_adj_mat(graph, max_n_vertices, num_edge_types, tie_fwd_bkwd=True):
    bwd_edge_offset = 0 if tie_fwd_bkwd else (num_edge_types // 2)
    amat = np.zeros((num_edge_types, max_n_vertices, max_n_vertices))
    # print(amat.shape)
    for src, e, dest in graph:
        # print("---------------")
        # print(src)
        # print(dest)

        amat[e-1, dest, src] = 1
        amat[e-1 + bwd_edge_offset, src, dest] = 1
    return amat

def _onehot(i, total):
    return [1.0 if j == i else 0.0 for j in range(total)]

   
class MonoLanguageProgramData():
   
    def __init__(self, opt, is_training=True, is_testing=False, live_test=False):

        self.node_dim = opt.node_dim
        self.state_dim = opt.state_dim

        self.is_training = is_training
        self.is_testing = is_testing

        if live_test:
       

            all_data_node_id, all_data_node_type = load_single_program(opt.test_graph_path)
            all_data_node_id = np.array(all_data_node_id)[0:len(all_data_node_id)]
            all_data_node_type = np.array(all_data_node_type)[0:len(all_data_node_type)]

        else:      
            base_name = os.path.basename(opt.path)
           
            if is_training:
                saved_input_filename = "%s/%s-%d-train.pkl" % (opt.path, base_name, opt.n_classes)
            if is_testing:
                saved_input_filename = "%s/%s-%d-test.pkl" % (opt.path, base_name, opt.n_classes)

            if os.path.exists(saved_input_filename): 
               input_file = open(saved_input_filename, 'rb')
               buf = input_file.read()
               all_data_node_id, all_data_node_type = pyarrow.deserialize(buf)
               input_file.close()
            else:
               all_data_node_id, all_data_node_type = load_program_graphs_from_directory(opt.path, is_training, is_testing, opt.n_classes)
               all_data_node_id = np.array(all_data_node_id)[0:len(all_data_node_id)]
               all_data_node_type = np.array(all_data_node_type)[0:len(all_data_node_type)]
               buf = pyarrow.serialize((all_data_node_id, all_data_node_type)).to_buffer()
               out = pyarrow.OSFile(saved_input_filename, 'wb')
               out.write(buf)
               out.close()
        
        self.pretrained_embeddings = opt.pretrained_embeddings
        self.batch_size = opt.train_batch_size

        label_lookup = {label: _onehot(label, opt.n_classes) for label in range(0, opt.n_classes)}

        self.label_lookup = label_lookup
        # if is_train == True:
        # print("Number of all data : " + str(len(all_data_node_id)))
        # else:
            # print("Number of all testing data : " + str(len(all_data_node_id)))
        self.n_edge_types =  find_max_edge_id(all_data_node_id)
        # print("Edge types : " + str(self.n_edge_types))
        max_node_id = find_max_node_id(all_data_node_id)
        min_node_id = find_min_node_id(all_data_node_id)
        # print("Max node id in data : " + str(max_node_id))
        # print("Min node id in data : " + str(min_node_id))
        max_node_type = find_max_node_id(all_data_node_type)
        min_node_type = find_min_node_id(all_data_node_type)
        # print("Max node type in data : " + str(max_node_type))
        # print("Min node type in data : " + str(min_node_type))
        # print("Max node id : " + str(max_node_id))
        # print("Max node type : " + str(max_node_type))
    
        self.n_node_by_id = max_node_id
        self.n_node_by_type = max_node_type
        
        all_data_node_id = convert_program_data(all_data_node_id)
        all_data_node_type = convert_program_data(all_data_node_type)

        self.all_data_node_id = all_data_node_id
        self.all_data_node_type = all_data_node_type

        self.data = self.process_raw_graphs()
    

    # ----- Data preprocessing and chunking into minibatches:
    def process_raw_graphs(self):
        # print("Processing raw graphs........")
        
        # Interesting idea to batch graphs with approximately similar size into the same bucket
        # The difference is that he original code from Microsoft batch exact similar size graphs into the same bucket
        bucket_sizes = np.array(list(range(65 , 2100 , 83)))
        

        buckets = defaultdict(list)
        x_dim = self.state_dim
        for i in range(len(self.all_data_node_id)):
            # print("--------------")
            graph = self.all_data_node_id[i][0]
            label = self.all_data_node_id[i][1]
            # print(label)

            label_one_hot = self.label_lookup[label-1]

            graph_node_type = self.all_data_node_type[i][0]
            

            # print(max([v for e in d['graph'] for v in [e[0], e[2]]]))
            chosen_bucket_idx = np.argmax(bucket_sizes > max([v for e in graph for v in [e[0], e[2]]]))
            chosen_bucket_size = bucket_sizes[chosen_bucket_idx]
            # print(chosen_bucket_size)
            n_active_nodes = self.state_dim

            num_nodes = find_num_nodes_of_graph(graph)

            # print("max node : " + str(num_nodes))
            node_inits = np.zeros([num_nodes+1, 30])

            # print(node_inits.shape)
            for k in range(len(graph)):
                src_node_type = graph_node_type[k][0]
                # print(src_node_type)
                tgt_node_type = graph_node_type[k][2]
                # print(node_type)
                src_node_id = graph[k][0]
                tgt_node_id = graph[k][0]
                # print("SRC node id : " + str(src_node_id))


                node_inits[int(src_node_id)] = self.pretrained_embeddings[int(src_node_type)]
                node_inits[int(tgt_node_id)] = self.pretrained_embeddings[int(tgt_node_type)]

            # adj_mat': graph_to_adj_mat(graph, chosen_bucket_size, self.n_edge_types, True),

            buckets[chosen_bucket_idx].append({
                "graph": graph,
                "init": node_inits,
                "labels": label_one_hot
            })

        #------------------------
        # To merge the bucket with too few elements together
        anchor_bucket_idx = None
        buckets_to_delete = []
        for bucket_idx, bucket_data in buckets.items():
            if len(bucket_data) > 0 and len(bucket_data) <=5:
                # print("anchor : " + str(anchor_bucket_idx))
                if anchor_bucket_idx is None:
                    # print("None , choose bukcket id : " + str(bucket_idx))
                    anchor_bucket_idx = bucket_idx               
                if anchor_bucket_idx != bucket_idx:
                    buckets[anchor_bucket_idx].extend(bucket_data)
                    buckets_to_delete.append(bucket_idx)

        for b in buckets_to_delete:
            buckets.pop(b,None)
        #------------------------

        total = 0
        for k,v in buckets.items():
            # print("--------")
            # print(k)
            # print(len(v))
            total = total + len(v)
        

        bucket_at_step = [[bucket_idx for _ in range(len(bucket_data))]
                          for bucket_idx, bucket_data in buckets.items()]

        # print("Bucket at step : " + str(bucket_at_step))
        # print(len(bucket_at_step))
        bucket_at_step = [x for y in bucket_at_step for x in y]

        #--------------------------------
            

        return (buckets, bucket_sizes, bucket_at_step)

    # Padding so the a batch has the same number of nodes
    def pad_batch(self, initial_node_representations, max_node):
        # remaining = max_node - len(initial_node_representations[0])
        for i in range(0,len(initial_node_representations)):
            remaining = max_node - len(initial_node_representations[i])
            initial_node_representations[i] = np.concatenate((initial_node_representations[i],np.zeros([remaining,30])), axis=0)

        return initial_node_representations

    def pad_annotations(self, initial_node_representations):
        return  np.pad(initial_node_representations,
                       pad_width=[[0, 0], [0, 0], [0, self.node_dim - self.state_dim]],
                       mode='constant')


    def make_batch(self, elements):
        batch_data = {'adjacency_matrix': [], 'init': [], 'labels': []}

        # find graph which has the largest number of nodes in batch
        max_node = find_num_nodes_of_graph(elements[0]["graph"])
        for d in elements:
            num_node = find_num_nodes_of_graph(d["graph"])
            if max_node < num_node:
                max_node = num_node
        # print("max node : " + str(max_node))

        for d in elements:
            adjacency_matrix = graph_to_adj_mat(d["graph"], max_node + 1, self.n_edge_types, True)
            batch_data['adjacency_matrix'].append(adjacency_matrix)
            batch_data['init'].append(d['init'])
            batch_data['labels'].append(d['labels'])

        batch_data["init"] = self.pad_batch(batch_data['init'], max_node + 1)
     
        batch_data['adjacency_matrix'] = np.array(batch_data['adjacency_matrix'])[0:len(batch_data['adjacency_matrix'])]
        batch_data['init'] = np.array(batch_data['init'])[0:len(batch_data['init'])]
        
        batch_data['labels'] = np.array(batch_data['labels'])[0:len(batch_data['labels'])]

        # print("adj shape : " + str(batch_data['adjacency_matrix'].shape))
        # print("init shape : " + str(batch_data['init'].shape))

        return batch_data, max_node

    def make_minibatch_iterator(self):
        (buckets, bucket_sizes, bucket_at_step) = self.data
        bucket_counters = defaultdict(int)
        

        if self.is_training:
            # print("Shuffling training data...........")
            # np.random.shuffle(bucket_at_step)
            for _, buckets_data in buckets.items():
                np.random.shuffle(buckets_data)

        for bucket_idx, bucket_data in buckets.items():

            elements = []
            samples = 0
            for i, element in enumerate(bucket_data):
                elements.append(element)
                samples += 1
                if (samples >= self.batch_size) or ((i == len(bucket_data)-1)):
                    batch_data, batch_max_node = self.make_batch(elements)
                    num_graphs = len(batch_data['init'])
                    initial_representations = batch_data['init']
                    initial_representations = self.pad_annotations(initial_representations)
                    # print(initial_representations.shape)
                    batch = {
                        "initial_representations": initial_representations,
                        "num_graphs": num_graphs,
                        "num_vertices": batch_max_node + 1,
                        "adjacency_matrix": batch_data['adjacency_matrix'],
                        "labels": batch_data['labels']
                    }

                    yield batch
                    elements = []
                    samples = 0

        # for step in range(len(bucket_at_step)):
        #     # print("-------------")
        #     bucket = bucket_at_step[step]
        #     start_idx = bucket_counters[bucket] * self.batch_size
        #     end_idx = (bucket_counters[bucket] + 1) * self.batch_size
        #     elements = buckets[bucket][start_idx:end_idx]

        #     batch_data, batch_max_node = self.make_batch(elements)
        #     num_graphs = len(batch_data['init'])
        #     initial_representations = batch_data['init']

        #     batch = {
        #         "initial_representations": initial_representations,
        #         "num_graphs": num_graphs,
        #         "num_vertices": batch_max_node + 1,
        #         "adjacency_matrix": batch_data['adjacency_matrix'],
        #         "labels": batch_data['labels']
        #     }


        #     yield batch


    def __len__(self):
        return len(self.all_data_node_id)



from setuptools import setup

setup(
    name='ggnn',
    version='0.0.7-7',
    license='Apache 2.0',
    author='Yijun Yu',
    author_email='y.yu@open.ac.uk',
    url='https://bitbucket.org/yijunyu/fast',
    long_description=('Generating Gated Graph Neural Networks from flat AST'
                      'serialization format.'),
    packages=['utils', 'utils.data', 'ggnn'],
    include_package_data=True,
    entry_points={'console_scripts': [
        "live_test=ggnn.live_test:main"]},
    install_requires=['flatast', 'tqdm'],
    requires=['flatast', 'tqdm', 'numpy', 'tensorboardX', 'sklearn', 'torch', 'pycurl'],
    description='Gated Graph Neural Networks',
)



import argparse
import random

import gzip
import pycurl
import pickle

import tensorflow as tf
from utils.data.dataset import MonoLanguageProgramData
from utils.data.dataset import load_single_program
from utils.utils import ThreadedIterator
from utils.utils import scale_attention_score_by_group
from utils.dense_ggnn import DenseGGNNModel
import os
import sys

import operator
import numpy as np
from sklearn.metrics import classification_report, confusion_matrix, accuracy_score

def generate_visualization(pb_path, attention_path):
    # attention_path = os.path.join(pb_path.split(".")[0] + ".csv")
    normal_html_path = os.path.join(pb_path.split(".")[0] + ".html")
    normal_node_id_cmd = "fast -H 0 -t -x " + attention_path + " " + pb_path  + " > " + normal_html_path
    os.system(normal_node_id_cmd)
    return normal_html_path

def generate_graph_files(opt, path):
    fbs_cmd = "fast -S -G " + path + " " + path.split(".")[0] + ".fbs"
    ggnn_cmd = "ggnn " + path.split(".")[0] + ".fbs" + " " + path.split(".")[0] + "_train.txt" + " " + path.split(".")[0] + ".txt"

    # print(ggnn_cmd)
    os.system(fbs_cmd)
    os.system(ggnn_cmd)

    opt.test_graph_path = path.split(".")[0] + ".txt"

def generate_pb(src_path):
    # print("Generating pb with src_path : " + src_path)
    pb_path = os.path.join(src_path.split(".")[0] + ".pb")
    # if not os.path.exists(src_path):
    cmd = "fast -p " + src_path + " " + pb_path
    os.system(cmd)
    return pb_path

def generate_files(opt, path):
    generate_graph_files(opt, path)
    pb_path = generate_pb(path)
    opt.pb_path = pb_path

def generate_attention_scores(opt, attention_scores):
    attention_score_map = {}
    for i, score in enumerate(attention_scores):
        attention_score_map[i] = float(score)

    attention_score_sorted = sorted(attention_score_map.items(), key=operator.itemgetter(1))
    attention_score_sorted.reverse()

    node_ids = []
    attention_score = []
    for element in attention_score_sorted:
        node_ids.append(element[0])
        attention_score.append(element[1])

    attention_score_scaled = scale_attention_score_by_group(attention_score)

    attention_score_scaled_map = {}
    for i, score in enumerate(attention_score_scaled):
        key = str(node_ids[i])
        attention_score_scaled_map[key] = float(score)

    attention_path = os.path.join(opt.test_file.split(".")[0] + ".csv")
    if os.path.exists(attention_path):
        os.remove(attention_path)

    with open(attention_path,"a") as f:
        for k, v in attention_score_scaled_map.items():
            f.write(str(k) + "," + str(v))
            f.write("\n")

    return attention_path

def fetch_data_from_github(filename):
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))
    fp = open(filename, "wb") 
    curl = pycurl.Curl()
    curl.setopt(pycurl.URL, os.path.join("https://raw.githubusercontent.com/bdqnghi/ggnn.tensorflow/master", filename))
    curl.setopt(pycurl.WRITEDATA, fp)
    curl.perform()
    curl.close()
    fp.close()   

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--workers', type=int, help='number of data loading workers', default=2)
    parser.add_argument('--train_batch_size', type=int, default=10, help='input batch size')
    parser.add_argument('--test_batch_size', type=int, default=5, help='input batch size')
    parser.add_argument('--state_dim', type=int, default=30, help='GGNN hidden state dimension size')
    parser.add_argument('--node_dim', type=int, default=200, help='node dimension size')
    parser.add_argument('--hidden_layer_size', type=int, default=200, help='size of hidden layer')
    parser.add_argument('--num_hidden_layer', type=int, default=1, help='number of hidden layer')
    parser.add_argument('--n_steps', type=int, default=10, help='propogation steps number of GGNN')
    parser.add_argument('--lr', type=float, default=0.001, help='learning rate')
    parser.add_argument('--cuda', action='store_true', help='enables cuda')
    parser.add_argument('--verbal', type=bool, default=True, help='print training info or not')
    parser.add_argument('--manualSeed', type=int, help='manual seed')
    parser.add_argument('--test_file', default="program_data/test_data/5/100_dead_code_1.java", help="test program")
    parser.add_argument('--n_classes', type=int, default=10, help='manual seed')
    parser.add_argument('--path', default="program_data/github_java_sort_function_babi", help='program data')
    parser.add_argument('--model_path', default="model/sum_softmax_hidden_layer_size_200_num_hidden_layer_1_node_dim_200", help='path to save the model')
    parser.add_argument('--n_hidden', type=int, default=50, help='number of hidden layers')
    parser.add_argument('--size_vocabulary', type=int, default=59, help='maximum number of node types')
    parser.add_argument('--log_path', default="logs/" ,help='log path for tensorboard')
    parser.add_argument('--aggregation', type=int, default=1, choices=range(0,4), help='0 for max pooling, 1 for attention with sum pooling, 2 for attention with max pooling, 3 for attention with average pooling')
    parser.add_argument('--distributed_function', type=int, default=0, choices=range(0,2), help='0 for softmax, 1 for sigmoid')
    parser.add_argument('--pretrained_embeddings_url', default="embedding/fast_pretrained_vectors.pkl.gz", help='pretrained embeddings url, there are 2 objects in this file, the first object is the embedding matrix, the other is the lookup dictionary')
    parser.add_argument('argv', nargs="+", help='filenames')
    opt = parser.parse_args()

    if len(opt.argv) == 1:
        opt.test_file = opt.argv[0]
    # Create model path folder if not exists
    if not os.path.exists(opt.model_path):
        os.makedirs(opt.model_path)
    
    generate_files(opt, opt.test_file)

    if not os.path.exists(opt.pretrained_embeddings_url):
        fetch_data_from_github(opt.pretrained_embeddings_url)
    with gzip.open(opt.pretrained_embeddings_url, 'rb') as fh:
        embeddings, embed_lookup = pickle.load(fh,encoding='latin1')
        opt.pretrained_embeddings = embeddings
        opt.pretrained_embed_lookup = embed_lookup

    checkfile = os.path.join(opt.model_path, 'cnn_tree.ckpt')    
    for f in ['checkpoint', 'cnn_tree.ckpt.index', 'cnn_tree.ckpt.meta', 'cnn_tree.ckpt.data-00000-of-00001']:
        filename = os.path.join(opt.model_path, f)
        if not os.path.exists(filename):
            fetch_data_from_github(filename)

    ckpt = tf.train.get_checkpoint_state(opt.model_path)
    
    test_dataset = MonoLanguageProgramData(opt, False, False, True)
    #opt.n_edge_types = test_dataset.n_edge_types
    opt.n_edge_types = 7
    ggnn = DenseGGNNModel(opt)

    # For debugging purpose
    nodes_representation = ggnn.nodes_representation
    graph_representation = ggnn.graph_representation
    logits = ggnn.logits
    softmax_values = ggnn.softmax_values
    attention_scores = ggnn.attention_scores

    saver = tf.train.Saver(save_relative_paths=True, max_to_keep=5)
    init = tf.global_variables_initializer()
   
    with tf.Session() as sess:
        sess.run(init)

        if ckpt and ckpt.model_checkpoint_path:
            # print("Continue training with old model")
            # print("Checkpoint path : " + str(ckpt.model_checkpoint_path))
            saver.restore(sess, ckpt.model_checkpoint_path)
            for i, var in enumerate(saver._var_list):
                print('Var {}: {}'.format(i, var))

        correct_labels = []
        predictions = []
        # print('Computing training accuracy...')
      
        batch_iterator = ThreadedIterator(test_dataset.make_minibatch_iterator(), max_queue_size=5)
        for step, batch_data in enumerate(batch_iterator):
            # print(batch_data["labels"])

            softmax_values_data, attention_scores_data = sess.run(
                [softmax_values, attention_scores],
                feed_dict={
                    ggnn.placeholders["initial_node_representation"]: batch_data["initial_representations"],
                    ggnn.placeholders["num_vertices"]: batch_data["num_vertices"],
                    ggnn.placeholders["adjacency_matrix"]:  batch_data['adjacency_matrix'],
                    ggnn.placeholders["labels"]:  batch_data['labels']
                }
            )

            print(softmax_values_data)
            # print(attention_scores_data)
            # print(len(attention_scores_data[0]))
            
            correct_labels.extend(np.argmax(batch_data['labels'],axis=1))
            predictions.extend(np.argmax(softmax_values_data,axis=1))


        # print("Num target : " + str(len(correct_labels)))
        print(correct_labels[0])
        print(predictions[0])
    
        attention_path = generate_attention_scores(opt, attention_scores_data[0])
        # print(attention_path)
        # print(opt.pb_path)
        generate_visualization(opt.pb_path,attention_path)

    # print(train_dataset.bucketed)

if __name__ == "__main__":
    main()
